/**
 ****************************************************************************************
 *
 * @file alarm_mechanism.c
 *
 * @brief Alarm mechanism implementation
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#include "osal.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "alarm_mechanism.h"
#include "sw_rtc_date_time.h"


/* Handle of OS timer */
__RETAINED_RW static OS_TIMER timer_h = NULL;

/* Hold the user-defined callback function */
__RETAINED_RW static alarm_cb cb = NULL;

/* Hold the current alarm settings */
__RETAINED_RW static time_alarm_t *alarm_cfg = NULL;



/*
 * OS timer callback function. This callback will be called
 * upon timer's expiration.
 */
static void timer_h_cb(OS_TIMER xTimer)
{
        // Check whether a user-defined function has been declared
        if (cb) {
                cb();
        }
}


/*
 * Initialize the ALARM mechanism. An OS timer is being created,
 * without enabling it.
 */
void time_alarm_init(void)
{
        /* Allocate memory for the alarm structure */
        alarm_cfg = (time_alarm_t *) OS_MALLOC(sizeof(time_alarm_t));
        OS_ASSERT(alarm_cfg);

        /* Clear the alarm structure */
        memset((uint8_t *)alarm_cfg, 0x00, sizeof(time_alarm_t));

        /*
         * Create one-shot OS timer
         *
         * \note  Timer period - expressed in OS ticks -
         *        must be greater than 1, otherwise an assertion is issued!
         */
        timer_h = OS_TIMER_CREATE("ALARM", OS_MS_2_TICKS(10),
                                        OS_TIMER_FAIL, NULL, timer_h_cb);
        OS_ASSERT(timer_h);

        /* Do not start timer here! */
}


/*
 * Function responsible for registering user-defined callback function
 */
void time_alarm_register_cb(alarm_cb user_cb)
{
        cb = user_cb;
}



/*
 * Function responsible for deleting  user-defined callback function
 */
void time_alarm_unregister_cb(void)
{
        cb = NULL;
}


/* Perform sanity checks in the alarm structure */
int time_alarm_sanity_check(time_alarm_t *val)
{
        if ((val->date.month_date > 31) || (val->date.month_date == 0) || (val->date.month > 12) ||
                 (val->date.month == 0) || (val->time.hour > 23) || (val->time.min > 59) || (val->time.sec > 59)) {
                return -1;
        } else {
                return 0;
        }
}


/* Function prototype */
uint32_t convert_months_to_sec(uint32_t month, bool leap_year);


#define DAYS_31_TO_SEC 2678400UL
#define DAYS_30_TO_SEC 2592000UL
#define DAYS_29_TO_SEC 2505600UL
#define DAYS_28_TO_SEC 2419200UL


/*
 * Given the current month, calculate how many seconds has elapsed
 * from the beginning of the year.
 */
uint32_t convert_months_to_sec(uint32_t month, bool leap_year)
{
        uint32_t month_sec;

        switch(month) {

        case JANUARY:
                month_sec = 2678400UL;
                break;
        case FEBRUARY:
                if (leap_year) {
                        month_sec = 2505600UL;
                } else {
                        month_sec = 2419200UL;
                }
                break;
        case MARCH:
                if (leap_year) {
                        month_sec = 7862400UL;
                } else {
                        month_sec = 7776000UL;
                }
                break;
        case APRIL:
                if (leap_year) {
                        month_sec = 10454400UL;
                } else {
                        month_sec = 10368000UL;
                }
                break;
        case MAY:
                if (leap_year) {
                        month_sec = 13132800UL;
                } else {
                        month_sec = 13046400UL;
                }
                break;
        case JUNE:
                if (leap_year) {
                        month_sec = 15724800UL;
                } else {
                        month_sec = 15638400UL;
                }
                break;
        case JULY:
                if (leap_year) {
                        month_sec = 18403200UL;
                } else {
                        month_sec = 18316800UL;
                }
                break;
        case AUGUST:
                if (leap_year) {
                        month_sec = 21081600UL;
                } else {
                        month_sec = 20995200UL;
                }
                break;
        case SEPTEMBER:
                if (leap_year) {
                        month_sec = 23673600UL;
                } else {
                        month_sec = 23587200UL;
                }
                break;
        case OCTOBER:
                if (leap_year) {
                        month_sec = 26352000UL;
                } else {
                        month_sec = 26265600UL;
                }
                break;
        case NOVEMBER:
                if (leap_year) {
                        month_sec = 28944000UL;
                } else {
                        month_sec = 28857600UL;
                }
                break;
        case DECEMBER:
                if (leap_year) {
                        month_sec = 31622400UL;
                } else {
                        month_sec = 31536000UL;
                }
                break;
        default:
                OS_ASSERT(0); // Invalid value!
        }

        return month_sec;
}



/* Check whether the given year is a leap year */
bool time_alarm_is_leap_year(uint32_t year)
{
        /*
         * Check if the year is a leap year, taking into consideration
         * the following three criteria:
         * 1. the year can be evenly divided by 4
         * 2. the year is divisible by 400
         * 3. the year is not divisible by 100
         */
        if ((!(year % 4)) ||
                ((!(year % 400)) && (year % 100))) {
                return true;
        } else {
                return false;
        }
}


/*
 * Convert date/time to seconds. It's OK to work with 32-bit values
 *
 * \note Intended for date/time acquired by the RTC mechanism
 *
 */
uint64_t convert_time_to_sec_rtc(sw_rtc_date_time_t *val)
{

        uint64_t cal  = 0;
        uint64_t total_sec = 0;

        bool leap_year = false;

        /* Convert seconds to seconds */
        cal = (uint64_t)val->time.sec;
        total_sec += cal;

        /* Convert minutes to seconds */
        cal = (uint64_t)(val->time.min * 60UL);
        total_sec += cal;

        /* Convert hours to seconds */
        cal = (uint64_t)(val->time.hour * 3600UL);
        total_sec += cal;

        if (val->date.month_date > 1) {
                /* Convert days of a month to seconds */
                cal = (uint64_t)((val->date.month_date - 1) * 86400UL);
                total_sec += cal;
        }

        if (val->date.month > 1) {
                /* Check if the year is a leap year */
                leap_year = time_alarm_is_leap_year(val->date.year);

                /* Convert months to seconds */
                cal = (uint64_t)convert_months_to_sec((val->date.month - 1), leap_year);
                total_sec += cal;
        }


        return total_sec;
}

/*
 * Convert date/time to seconds. It's OK to work with 32-bit values.
 *
 * \note Intended for date/time acquired by the alarm mechanism
 *
 */
uint64_t convert_time_to_sec_alarm(time_alarm_t *val)
{
        uint64_t cal  = 0;
        uint64_t total_sec = 0;

        bool leap_year = false;

        /* Convert seconds to seconds */
        cal = (uint64_t)val->time.sec;
        total_sec += cal;

        /* Convert minutes to seconds */
        cal = (uint64_t)(val->time.min * 60UL);
        total_sec += cal;

        /* Convert hours to seconds */
        cal = (uint64_t)(val->time.hour * 3600UL);
        total_sec += cal;

        if (val->date.month_date > 1) {
                /* Convert days of a month to seconds */
                cal = (uint64_t)((val->date.month_date - 1) * 86400UL);
                total_sec += cal;
        }

        if (val->date.month > 1) {
                /* Check if the year is a leap year */
                leap_year = time_alarm_is_leap_year(val->date.year);

                /* Convert months to seconds */
                cal = (uint64_t)convert_months_to_sec((val->date.month - 1), leap_year);
                total_sec += cal;
        }

        return total_sec;
}


/*
 * Function responsible for setting an alarm event.
 */
int time_alarm_set(time_alarm_t *val)
{
        OS_BASE_TYPE ret = 0;

        int error;


        /**
         * Perform sanity checks on the provided values.
         *
         * \note If the returned value is -1 then the provided values are out of boundaries
         *
         */
        error = time_alarm_sanity_check(val);
        OS_ASSERT(error == 0);


        /*
         * Allocate memory for the date/time structure
         */
        sw_rtc_date_time_t *date_time = (sw_rtc_date_time_t *) OS_MALLOC(sizeof(sw_rtc_date_time_t));
        OS_ASSERT(date_time);


        /* Read the current date/time using the SW RTC mechanism */
        sw_rtc_date_time_read(date_time);


        /*
         * It's not allowed to set ALARM events referred to past or set
         * to hit in more than 1 year (given the current date/time) due
         * to uint32_t invoked numbers.
         */
        uint32_t year_diff = (val->date.year - date_time->date.year);
        if (year_diff > 1) {
                return -1;
        }


        /* Convert the current date/time into seconds */
        uint64_t total_sec_rtc = convert_time_to_sec_rtc(date_time);

        /* Convert the alarm parameters into seconds */
        uint64_t total_sec_alarm = convert_time_to_sec_alarm(val);

        uint64_t total_sec_dif;

        /* Compute the time interval that an ALARM event should hit (based on the current date/time)*/
        if (val->date.year > date_time->date.year) {
                /* Check if the year is a leap year */
                if (time_alarm_is_leap_year(date_time->date.year)) {

                        total_sec_rtc = ((uint64_t)31622400UL - total_sec_rtc);
                        total_sec_dif = total_sec_alarm + total_sec_rtc;
                } else {

                        total_sec_rtc = ((uint64_t)31536000UL - total_sec_rtc);
                        total_sec_dif = total_sec_alarm + total_sec_rtc;
                }

        } else {
               total_sec_dif = total_sec_alarm - total_sec_rtc;
        }


        /* Convert seconds to milliseconds */
        total_sec_dif *= (uint64_t)1000;

        /* Convert milliseconds into OS ticks (based on the lp clock used) */
        uint64_t timer_ticks = ((total_sec_dif * configTICK_RATE_HZ) / (uint64_t)1000);

        /* Perform sanity checks */
        if (timer_ticks >= (uint64_t)4294967295UL) {
              return -1;
        }

        /*
         * Start the OS timer with the new period
         *
         * \note: Changing the period of a dormant timer will also start the timer!
         */
        ret = OS_TIMER_CHANGE_PERIOD(timer_h, (uint32_t)timer_ticks, OS_TIMER_FOREVER);
        OS_ASSERT(ret == OS_OK);


        /* Free the previously allocated memory for date/time. */
        OS_FREE(date_time);

        /* Copy the current configurations */
        memcpy((uint8_t *)alarm_cfg, (uint8_t *)val, sizeof(time_alarm_t));


        return 0;
}


/* Function responsible for clearing any previously configured alarm event */
void time_alarm_clear(void)
{
        /* Clear the associated variables */
        memset((uint8_t *)alarm_cfg, 0x00, sizeof(time_alarm_t));


        /* Stop the associated OS timer */
        OS_TIMER_STOP(timer_h, OS_TIMER_FOREVER);
}
