/**
 ****************************************************************************************
 *
 * @file alarm_mechanism.h
 *
 * @brief Alarm mechanism APIs
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#ifndef ALARM_MECHANISM_H_
#define ALARM_MECHANISM_H_


/* Date related parameters */
typedef struct {
        uint32_t year;           // Current year: there is no limitation
        uint32_t month;          // Range: (1..12)
        uint32_t month_date;     // Range: (1..31) or (1..30) or (1..29) or (1..28)
} time_alarm_date_t;


/* Time related parameters */
typedef struct {
        uint32_t hour;  // Range: (0..23)
        uint32_t min;   // Range: (0..59)
        uint32_t sec;   // Range: (0..59)
} time_alarm_time_t;


/**
 *
 * \brief Alarm parameters
 *
 */
typedef struct time_alarm {
        time_alarm_date_t date;
        time_alarm_time_t time;
} time_alarm_t;




/**
 *
 * \brief User-defined callback function
 *
 * This function will be triggered upon an alarm event
 */
typedef void (*alarm_cb) (void);




/*
 * \brief Alarm mechanism initialization.
 *
 * This function initializes all the resources required for the alarm mechanism.
 *
 * \note It should be the first invoked API before any other alarm related operation.
 *
 * \note The code supports the configuration of one alarm event!
 *
 */
void time_alarm_init(void);


/*
 * \brief Callback function registration
 *
 * This function should be used for registering a callback function,
 * that will be triggered upon an alarm event.
 *
 * \param [in] user_cb  The user-defined callback function.
 *
 *
 * \note If \p user_cb is set to NULL, none callback function will be called upon an alarm event.
 *
 */
void time_alarm_register_cb(alarm_cb user_cb);



/**
 *
 * \brief Callback function deletion
 *
 * This function should be used for unregistering any previously registered
 * callback function.
 *
 */
void time_alarm_unregister_cb(void);


/*
 * \brief Set alarms.
 *
 * This function should be used for setting alarms. It performs an RTC read operation to acquire
 * the current date/time and based on that time reference, it configures an alarm event.
 *
 *
 * \param [in] alarm  The date/time that an alarm should be triggered (based on the current date/time)
 *
 *
 * \return 0 if succeeded, otherwise -1.
 *
 *
 * \note The code performs sanity checks to identify values out of boundaries.
 *       For any invalid value, the code gets stuck in assertions!
 *
  * \note: Due to 32-bit invoked numbers, the max. possible time interval
  *        is approx. 99 days. In other words, an alarm  can be triggered
  *        max. 99 days after the current date and time.
 *
 *        2^32 = 4294967295 * 1,95 ms (OS tick granularity) = 8589934592 ms or (8589934592 / 1000) = 8589935 sec
 *        (8589935 / 86400 (per day)) = approx. 99 days
 *
 *
 * \warning  The SW RTC mechanism should be up and running before invoking this API!!!
 *
 *
 * \warning: It's user's responsibility to set a valid alarm. For instance: do not configure an
 *           alarm that is referred to a past time (compared to the current date/time).
 *
 */
int time_alarm_set(time_alarm_t *alarm);


/*
 * \brief Clear an alarm event
 *
 * This function should be used for disabling any previously configured alarm event.
 *
 */
void time_alarm_clear(void);


#endif /* ALARM_MECHANISM_H_ */
