/**
 ****************************************************************************************
 *
 * @file alarm_task.c
 *
 * @brief Alarm application task
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#include <string.h>
#include <stdio.h>

#include "alarm_mechanism.h"
#include "sw_rtc_date_time.h"
#include "osal.h"

#include <stdbool.h>


/******** Notification Bitmasks *************/
#define TIME_ALARM_TRIG_NOTIF   (1 << 1)


/* Handle of [prvALARM_Task] task */
__RETAINED_RW static OS_TASK alarm_task_h = NULL;


/*
 * Macro for enabling/disabling debugging messages on the serial console.
 * 1: enables  debugging messages
 * 0: disables debugging messages
 */
#define TIME_ALARM_DBG_MESSAGES_CONSOLE    1



// ************************************************ Before using the alarm mechanism make sure that the RTC mechanism is up and running ******************************************************


/*
 * Structure that holds alarm related parameters (preferred data/time)
 *
 * \note: The application supports the configuration of one alarm event.
 *
 * \warning: It's user's responsibility to set a valid alarm event. For instance: do not set an
 *           alarm event that is referred to past (given the current date/time)
 *
 * \warning: Due to 32-bit invoked numbers the maximum possible time interval is approx. 99 days. In other words,
 *           an alarm event can be triggered maximum after 99 days (given the current date and time)
 *
 */
__RETAINED_RW time_alarm_t alarm_cfg = {
        /*
         * Declare all the date related info
         *
         * \note The code performs sanity checks to identify values out of boundaries.
         *
         */
        .date = {
                .year       = 2018,
                .month      = 10,
                .month_date = 1,
        },
        /* Declare all the time related info */
        .time = {
                .hour       = 0,
                .min        = 1,
                .sec        = 0,
        }

};



/*
 * User-defined callback function, called upon an alarm event.
 * It should have that specific prototype!
 */
void my_alarm_cb(void)
{
        OS_TASK_NOTIFY(alarm_task_h, TIME_ALARM_TRIG_NOTIF,
                                                OS_NOTIFY_SET_BITS);
}

/*
 * Task responsible for performing alarm related operations
 * */
void prvALARM_Task( void *pvParameters )
{
        int error = 0;

        /* Get the handle of the current task */
        alarm_task_h = OS_GET_CURRENT_TASK();

        /*
         * Before proceed to any ALARM functionality, make sure that
         * the SW RTC mechanism is ready and configured.
         *
         * 0: SW RTC mechanism uninitialized
         * 1: SW RTC mechanism initialized
         *
         */
        while (!sw_rtc_check_status()) ;

        /*
         * Initialize the alarm mechanism. This function should be invoked
         * before any other related API.
         */
        time_alarm_init();


        /*
         * Register callback function that will be called upon an alarm event.
         */
        time_alarm_register_cb(my_alarm_cb);


        /*
         * Set an alarm event. This function returns '0' in case of
         * successful alarm event configuration, -1 otherwise.
         *
         * \warnig Before setting alarm events, make sure that the RTC mechanism is up and running!!!
         *
         */
        error = time_alarm_set(&alarm_cfg);

        if (error == 0) {
                printf("\n\rValid ALARM configurations\n\r");
        } else {
                printf("\n\rInvalid ALARM configurations\n\r");
        }


        for (;;) {

                OS_BASE_TYPE ret;
                uint32_t notif;

                /*
                 * Block forever waiting for any of the notification bits,
                 * then clear them all
                 */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS,
                                                &notif, OS_TASK_NOTIFY_FOREVER);


                /*
                 * Thus, the return value must always be OS_OK
                 */
                OS_ASSERT(ret == OS_OK);

                /* This code block is executed upon an alarm event */
                if (notif & TIME_ALARM_TRIG_NOTIF) {

                        if (TIME_ALARM_DBG_MESSAGES_CONSOLE) {
                                printf("Alarm has been triggered successfully\n\r");
                        }
                }

        }
}
