/**
 ****************************************************************************************
 *
 * @file alert_mechanism.c
 *
 * @brief Alert mechanism implementation
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#include "alert_mechanism.h"

#include <string.h>
#include "osal.h"


/*
 * Masks for positioning alert sources (alert class)
 */
#define ALERT_SOURCE_HOURS_Pos   (0x00)
#define ALERT_SOURCE_MINUTES_Pos (0x01)
#define ALERT_SOURCE_SECONDS_Pos (0x02)


/* Handles of OS timers - One for each ALERT class */
__RETAINED_RW static OS_TIMER timer_h = NULL;
__RETAINED_RW static OS_TIMER timer_m = NULL;
__RETAINED_RW static OS_TIMER timer_s = NULL;



/*
 * Hold the current user-defined callback function
 */
__RETAINED_RW static alert_cb cb = NULL;

/*
 * Hold the current alert configurations.
 */
__RETAINED_RW static time_alert_t *alert_cfg = NULL;



/*
 * OS timer callback, dedicated for Hours Alert Class.
 * This callback is called upon timer's expiration.
 */
static void timer_h_cb(OS_TIMER timer)
{
        // Check whether the [repeat] flag is set, then restart timer.
        if (alert_cfg->repeat_h == pdTRUE) {
                OS_TIMER_START(timer, OS_TIMER_FOREVER);
        }

        // Check whether a user-defined function has been declared.
        if (cb) {
                cb((uint8_t)(1 << ALERT_SOURCE_HOURS_Pos));
        }

}


/*
 * OS timer callback, dedicated for Minutes Alert Class.
 * This callback is called upon timer's expiration.
 */
void timer_m_cb(OS_TIMER timer)
{
        // Check whether the [repeat] flag is set, then restart timer.
        if (alert_cfg->repeat_m == pdTRUE) {
                OS_TIMER_START(timer, OS_TIMER_FOREVER);
        }

        // Check whether a user-defined function has been declared.
        if (cb) {
                cb((uint8_t)(1 << ALERT_SOURCE_MINUTES_Pos));
        }

}



/*
 * OS timer callback, dedicated for Seconds Alert Class.
 * This callback is called upon timer's expiration.
 */
void timer_s_cb(OS_TIMER timer)
{
        // Check whether the [repeat] flag is set, then restart timer.
        if (alert_cfg->repeat_s == pdTRUE) {
                OS_TIMER_START(timer, OS_TIMER_FOREVER);
        }

        // Check whether a user-defined function has been declared.
        if (cb) {
                cb((uint8_t)(1 << ALERT_SOURCE_SECONDS_Pos));
        }

}




/*
 * Initialize the alert mechanism. Three OS timers are being created,
 * one for each ALERT class, without enabling them.
 */
void time_alert_init(void)
{
       /* Allocate memory for the alert parameters */
       alert_cfg = (time_alert_t *) OS_MALLOC(sizeof(time_alert_t));
       OS_ASSERT(alert_cfg);

       /* Clear the alert structure */
       memset((uint8_t *)alert_cfg, 0x00, sizeof(time_alert_t));


       /*
        * Create one-shot OS timer dedicated for Hours Alert Class.
        *
        * \note  Timer period - expressed in OS ticks -
        *        must be greater than 1, otherwise an assertion is issued!
        */
       timer_h = OS_TIMER_CREATE("HOURS", OS_MS_2_TICKS(10),
                                       OS_TIMER_FAIL, (void *)0, timer_h_cb);
       OS_ASSERT(timer_h);

       /*
        * Create one-shot OS timer dedicated for Minutes Alert Class.
        *
        * \note:  Timer's period - expressed in OS ticks -
        *         must be greater than 1, otherwise an assertion is issued!
        */
       timer_m = OS_TIMER_CREATE("MINUTES", OS_MS_2_TICKS(10),
                                        OS_TIMER_FAIL, (void *)0, timer_m_cb);
       OS_ASSERT(timer_m);

       /*
        * Create one-shot OS timer dedicated for Seconds Alert Class.
        *
        * \note: Timer's period - expressed in OS ticks -
        *        must be greater than 1, otherwise an assertion is issued.
        */
       timer_s = OS_TIMER_CREATE("SECONDS", OS_MS_2_TICKS(10),
                                         OS_TIMER_FAIL, (void *)0, timer_s_cb);
       OS_ASSERT(timer_s);

}


/*
 * Function responsible for setting an alert
 */
void time_alert_class_set(uint16_t value, bool repeat, SET_ALERT_CLASS alert)
{
        OS_BASE_TYPE ret = 0;
        uint64_t timer_ticks = 0;

        switch(alert) {

        /* Set alerts for Hours Alert Class */
        case SET_ALERT_CLASS_Hour:
                /* Then check whether the new value assigned is non-zero */
                if (value > 0) {
                        /*
                         * Recalculate timer period, that is,
                         * (hours * 60 minutes/hour * 60 seconds/minute * 1000 milliseconds/second)
                         *
                         * Note: Changing the period of a dormant timer will also
                         * start the timer!
                         */

                        timer_ticks = (((uint64_t)value * 3600000 * configTICK_RATE_HZ) / 1000);

                        /* Perform a sanity check for 32-bit max number */
                        if (timer_ticks >= (uint64_t)4294967295UL) {
                              OS_ASSERT(0);
                        }

                        ret = OS_TIMER_CHANGE_PERIOD(timer_h,
                                        (uint32_t)timer_ticks, OS_TIMER_FOREVER);

                        OS_ASSERT(ret == OS_OK);

                 /* Otherwise the new value is zero and hence timer should be halted */
                 } else {
                         ret = OS_TIMER_STOP(timer_h, OS_TIMER_FOREVER);
                         OS_ASSERT(ret == OS_OK);
                 }

                /* Store the new parameters for Hours Alert Class */
                alert_cfg->hour     = value;
                alert_cfg->repeat_h = repeat;
                break;

        /* Set alerts for Minutes Alert Class */
        case SET_ALERT_CLASS_Min:
                /* Then check whether the new value assigned is non-zero */
                if (value > 0) {
                        /*
                         * Recalculate timer period, that is,
                         * (minutes * 60 seconds/minute * 1000 milliseconds/second)
                         *
                         * Note: changing the period of a dormant timer will also
                         * start the timer!
                         */
                         timer_ticks = (((uint64_t)value * 60000 * configTICK_RATE_HZ) / 1000);

                         /* Perform a sanity check for 32-bit max number */
                         if (timer_ticks >= (uint64_t)4294967295UL) {
                               OS_ASSERT(0);
                         }

                         ret = OS_TIMER_CHANGE_PERIOD(timer_m,
                                          (uint32_t)timer_ticks, OS_TIMER_FOREVER);

                         OS_ASSERT(ret == OS_OK);

                  /* Otherwise the new value is zero and hence timer should be halted */
                  } else {
                          ret = OS_TIMER_STOP(timer_m, OS_TIMER_FOREVER);
                          OS_ASSERT(ret == OS_OK);
                        }

                /* Store the new parameters for Minutes Alert Class */
                alert_cfg->min = value;
                alert_cfg->repeat_m = repeat;
                break;

        /* Set alerts for Seconds Alert Class */
        case SET_ALERT_CLASS_Sec:
                /* Then check whether the new value assigned is non-zero */
                if (value > 0) {
                        /*
                         * Recalculate timer period, that is,
                         * (seconds * 1000 milliseconds/second)
                         *
                         * Note: changing the period of a dormant timer will also
                         * start the timer!
                         */

                         timer_ticks = (((uint64_t)value * 1000 * configTICK_RATE_HZ) / 1000);

                         /* Perform a sanity check for 32-bit max number */
                         if (timer_ticks >= (uint64_t)4294967295UL) {
                               OS_ASSERT(0);
                         }

                         ret = OS_TIMER_CHANGE_PERIOD(timer_s,
                                       (uint32_t)timer_ticks, OS_TIMER_FOREVER);

                         OS_ASSERT(ret == OS_OK);

                        /* Otherwise the new value is zero and hence timer should be halted */
                  } else {
                          ret = OS_TIMER_STOP(timer_s, OS_TIMER_FOREVER);
                          OS_ASSERT(ret == OS_OK);
                        }

                /* Store the new parameters for Seconds Alert Class */
                alert_cfg->sec = value;
                alert_cfg->repeat_s = repeat;
                break;

        default:
                OS_ASSERT(0);
        }

}


/*
 * Function responsible for settings alerts of all the available alert classes
 */
void time_alert_class_set_all(time_alert_t *alert)
{
        time_alert_class_set(alert->hour, alert->repeat_h, SET_ALERT_CLASS_Hour);
        time_alert_class_set(alert->min,  alert->repeat_m, SET_ALERT_CLASS_Min);
        time_alert_class_set(alert->sec,  alert->repeat_s, SET_ALERT_CLASS_Sec);
}


/*
 * Function responsible for clearing alerts
 */
void time_alert_class_clear(CLEAR_ALERT_CLASS alert)
{

        switch(alert) {

        /* Disable alerts for Hours Alert Class */
        case CLEAR_ALERT_CLASS_Hour:
                /* Clear the associated variables */
                alert_cfg->hour = 0;
                alert_cfg->repeat_h = false;

                /*Stop the associated OS timer */
                OS_TIMER_STOP(timer_h, OS_TIMER_FOREVER);
                break;

        /* Disable alerts for Minutes Alert Class */
        case CLEAR_ALERT_CLASS_Min:
                /* Clear the associated variables */
                alert_cfg->min = 0;
                alert_cfg->repeat_m = false;

                /*Stop the associated OS timer */
                OS_TIMER_STOP(timer_m, OS_TIMER_FOREVER);
                break;

        /* Disable alerts for Seconds Alert Class */
        case CLEAR_ALERT_CLASS_Sec:
                /* Clear the associated variables */
                alert_cfg->sec = 0;
                alert_cfg->repeat_s = false;

                /*Stop the associated OS timer */
                OS_TIMER_STOP(timer_s, OS_TIMER_FOREVER);
                break;

        /* Disable alerts for all classes */
        case CLEAR_ALERT_CLASS_All:
                /* Clear all the alert variables */
                memset((uint8_t *)alert_cfg, 0x00, sizeof(time_alert_t));

                /*Stop all the associated OS timers */
                OS_TIMER_STOP(timer_h, OS_TIMER_FOREVER);
                OS_TIMER_STOP(timer_m, OS_TIMER_FOREVER);
                OS_TIMER_STOP(timer_s, OS_TIMER_FOREVER);
                break;

        default:
                OS_ASSERT(0);
        }

}



/*
 * Function responsible for registering the user-defined callback function
 */
void time_alert_register_cb(alert_cb user_cb)
{
        cb = user_cb;
}

/*
 * Function responsible for deleting the user-defined callback function
 */
void time_alert_unregister_cb(void)
{
        cb = NULL;
}
