/**
 ****************************************************************************************
 *
 * @file alert_mechanism.h
 *
 * @brief Alert mechanism APIs
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#ifndef ALERT_MECHANISM_H_
#define ALERT_MECHANISM_H_

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>


/**
 *
 * \brief Alert related parameters.
 *
 * \note There are three distinct ALERT classes that can be configured individually.
 *
 */
typedef struct time_alert {
        /*
         * A value equal to zero means that the corresponding alert class is disabled.
         */
        uint16_t hour;
        uint16_t min;
        uint16_t sec;

        /*
         * Define whether an alert will be repeatable. Valid values are:
         * true: for repeatable alarms, false otherwise.
         */
        bool repeat_h;
        bool repeat_m;
        bool repeat_s;
} time_alert_t;


/**
 *
 * \brief Clear an alert class
 *
 */
typedef enum {
     CLEAR_ALERT_CLASS_Hour = 0,
     CLEAR_ALERT_CLASS_Min  = 1,
     CLEAR_ALERT_CLASS_Sec  = 2,
     CLEAR_ALERT_CLASS_All  = 3,
} CLEAR_ALERT_CLASS;


/**
 *
 * \brief Set an alert class
 *
 */
typedef enum {
        SET_ALERT_CLASS_Hour = 0,
        SET_ALERT_CLASS_Min  = 1,
        SET_ALERT_CLASS_Sec  = 2,
} SET_ALERT_CLASS;



/**
 *
 * \brief Alert Class Masks
 *
 * These masks should be used within callback functions to identify
 * the source of an alert event. This is necessary since all the
 * ALERT classes will call the same callback function.
 *
 */
#define ALERT_SOURCE_HOURS_Msk   (0x01)
#define ALERT_SOURCE_MINUTES_Msk (0x02)
#define ALERT_SOURCE_SECONDS_Msk (0x04)



/**
 *
 * \brief User-defined callback function
 *
 * This function will be called upon an alert event. It should have that specific prototype!
 *
 * \note All the alert classes will call the same callback. User is responsible to identify
 *       the source of an alert event. Following the recommended callback usage:
 *
 * \code{.c}
 * void my_alert_cb(uint8_t source)
 * {
 *      if (source & ALERT_SOURCE_HOURS_Msk) {
 *              // Do something...
 *       }
 *
 *       if (source & ALERT_SOURCE_MINUTES_Msk) {
 *              // Do something...
 *       }
 *
 *       if (source & ALERT_SOURCE_SECONDS_Msk) {
 *              // Do something...
 *       }
 * }
 * \endcode
 */
typedef void (*alert_cb)(uint8_t source);



/**
 *
 * \brief Alert mechanism initialization
 *
 * This function initializes all the resources required for the alert mechanism.
 *
 * \note It should be the first invoked API before any other alert related operation.
 *
 */
void time_alert_init(void);



/**
 *
 * \brief Callback function registration
 *
 *
 * This function should be used for registering a callback function, called upon an alert event.
 *
 *
 * \param [in] user_cb  The user-defined callback function.
 *
 * \note If \p user_cb is set to NULL, none callback function will be called upon an alert event.
 *
 */
void time_alert_register_cb(alert_cb user_cb);


/**
 *
 * \brief Callback function deletion
 *
 * This function should be used for unregistering any previously declared callback.
 *
 */
void time_alert_unregister_cb(void);



/**
 *
 * \brief Set alerts for a specific ALERT class
 *
 * This function should be used for declaring alerts of a specific ALERT class.
 * There are three distinct alert classes that can be configured individually.
 *
 * The following code demonstrates alert configurations, expressed in minutes. Assuming we
 * want alerts to be triggered every 5 minutes (triggered repeatedly).
 *
 * \code{.c}
 * time_alert_set_class(5, true, SET_ALERT_CLASS_Min)
 * \endcode
 *
 * The following code demonstrates alert configurations, expressed in seconds. Assuming we
 * want an alert after 5 minutes (triggered once)
 *
 * \code{.c}
 * time_alert_set_class(5, false, SET_ALERT_CLASS_Min)
 * \endcode
 *
 *
 * \param [in] value   Time interval of alert events.
 * \param [in] repeat  If set to false, alerts will be triggered once, repeatedly otherwise.
 * \param [in] alert   One of the three alert classes.
 *
 *
 * \note If \p value is set to zero, alerts for that specific alert class will be disabled.
 *
 */
void time_alert_class_set(uint16_t value, bool repeat, SET_ALERT_CLASS alert);


/**
 * \brief Set alerts for all all the three alert classes
 *
 * This is a shortcut function since it can configure all the three alert classes.
 *
 * \param [in] alert  Parameters for all the tree alert classes.
 *
 */
void time_alert_class_set_all(time_alert_t *alert);



/**
 *
 * \brief Clear alert events
 *
 * This function should be used for disabling alert events of an individual or all the three alert classes
 *
 * \param [in] alert  The alert class(es)
 *
 */
void time_alert_class_clear(CLEAR_ALERT_CLASS alert);



#endif /* ALERT_MECHANISM_H_ */
