/**
 ****************************************************************************************
 *
 * @file alert_task.c
 *
 * @brief Alert application task
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#include <string.h>
#include <stdio.h>

#include "alert_mechanism.h"
#include "osal.h"

#include <stdbool.h>

#include "sys_power_mgr.h"
#include "hw_led.h"
#include "hw_breath.h"

#include "sw_rtc_date_time.h"



/* Handle of [prvALERT_Task] task */
__RETAINED_RW OS_TASK alert_task_h = NULL;

/******** Notification Bitmasks *************/
#define TIME_ALERT_TRIG_NOTIF_H  (1 << 1)
#define TIME_ALERT_TRIG_NOTIF_M  (1 << 2)
#define TIME_ALERT_TRIG_NOTIF_S  (1 << 3)


/*
 * Macro for enabling/disabling debugging messages on the serial console:
 *
 * 1: enables  debugging messages
 * 0: disables debugging messages
 */
#define TIME_ALERT_DBG_MESSAGES_CONSOLE    1

/*
 * Macro for enabling/disabling blinking functionality upon an alert event:
 *
 * 1: enables  blinking functionality
 * 0: disables blinking functionality
 */
#define TIME_ALERT_BREATH_TIMER_EN         1



/*
 * Keep track of device state (active/sleep)
 */
__RETAINED_RW static bool old_alerting = false;

/* Function prototypes */
void time_alert_set_mode(bool new_alerting);



/*
 * Turn on LED1 on Pro DevKit (blinking functionality).
 * LED1 is driven by breath timer.
 */
void time_alert_led_blink_start(breath_config *breath_cfg)
{
        /* Force the device to stay in active mode */
        time_alert_set_mode(true);

        /* Configure the breath timer */
        hw_breath_init(breath_cfg);

        /* Configure the white LED1 */
        hw_led_set_led1_src(HW_LED_SRC1_BREATH);
        hw_led_enable_led1(true);

        /* Enable the breath timer */
        hw_breath_enable();
}


/* Turn off LED1 on Pro DevKit */
void time_alert_led_blink_stop(void)
{
        /* Resume device to its initial state */
        time_alert_set_mode(false);

        /* Disable breath timer */
        hw_breath_disable();
}


/*
 * Check whether the device should be forced to stay
 * awake or it's time to resume to its initial state.
 *
 * \note Number of calls to pm_stay_alive() must match the
 *       number of calls to pm_resume_sleep()!!!!!!!
 */
void time_alert_set_mode(bool new_alerting)
{
        /* First check if a new alert flag has been assigned */
        if (old_alerting != new_alerting) {

                /* If flag is set to true then force the device to stay alive */
                if (new_alerting) {
                        pm_stay_alive();

                /*
                 * Otherwise the new value is zero and thus the device should
                 * resume to its initial state.
                 */
                } else {
                        pm_resume_sleep();
                }

                /* Make a copy of the new alert flag */
                old_alerting = new_alerting;
        }
}


/*
 * Structure that holds alert related parameters
 *
 * \note: There are three distinct alert classes which can be
 *        enabled-disabled-configured individually.
 *
 */
__RETAINED_RW time_alert_t my_alert = {
        .hour     = 4,   // Set alerts expressed in hours   (every x hours  ).
        .min      = 2,   // Set alerts expressed in minutes (every x minutes).
        .sec      = 5,   // Set alerts expressed in seconds (every x seconds).

        .repeat_h = false,  // If set to false, alerts are triggered once, repeatedly otherwise.
        .repeat_m = true,   // If set to false, alerts are triggered once, repeatedly otherwise.
        .repeat_s = false,  // If set to false, alerts are triggered once, repeatedly otherwise.
};


/*
 * User-defined callback function, called upon an alert event.
 * It should have that specific prototype!
 *
 * \note: All the alert classes will call the same callback. The \p source parameter
 *        should be used to identify the source of the event.
 *
 */
void my_alert_cb(uint8_t source)
{

        /*
         * Check whether Hours Alert Class triggered the alert.
         * Then notify the main task accordingly.
         */
        if (source & ALERT_SOURCE_HOURS_Msk) {

                OS_TASK_NOTIFY(alert_task_h, TIME_ALERT_TRIG_NOTIF_H,
                                                          OS_NOTIFY_SET_BITS);
        }

        /*
         * Check whether Minutes Alert Class triggered the alert.
         * Then notify the main task accordingly.
         */
        if (source & ALERT_SOURCE_MINUTES_Msk) {

                OS_TASK_NOTIFY(alert_task_h, TIME_ALERT_TRIG_NOTIF_M,
                                                          OS_NOTIFY_SET_BITS);
        }

        /*
         * Check whether Seconds Alert Class triggered the alert.
         * Then notify the main task accordingly.
         */
        if (source & ALERT_SOURCE_SECONDS_Msk) {

                OS_TASK_NOTIFY(alert_task_h, TIME_ALERT_TRIG_NOTIF_S,
                                                          OS_NOTIFY_SET_BITS);
        }
}


/*
 * Task responsible for performing alert operations.
 */
void prvALERT_Task( void *pvParameters )
{

        /* Get the handler of the current task */
        alert_task_h = OS_GET_CURRENT_TASK();


        /*
         * Initialize the alert mechanism. This function should be invoked
         * before any other related API.
         */
        time_alert_init();


        /*
         * Register callback function that will be called upon an alert event.
         */
        time_alert_register_cb(my_alert_cb);


        /*
         * Set alerts, expressed in minutes.
         *
         * \note If repeat flag is set to true, alerts are triggered every x minutes
         *
         * \note If a new alert configuration is performed before the execution of the current alert,
         *       the old alert will be discarded and the new one will be taken into consideration.
         *
         * \note At any time, alerts can be disabled by calling time_alert_class_clear()
         *
         */
        time_alert_class_set(my_alert.min, my_alert.repeat_m, SET_ALERT_CLASS_Min);


        /*
         * Set alerts expressed in hours. Configure an alert to be triggered after 1 hour.
         *
         * \note If repeat flag is set to true, alerts are triggered every x minutes
         *
         * \note If a new alert configuration is performed before the execution of the current alert,
         *       the old alert will be discarded and the new one will be taken into consideration.
         *
         * \note At any time, alerts can be disabled by calling time_alert_class_clear()
         *
         */
        time_alert_class_set(my_alert.hour, my_alert.repeat_h, SET_ALERT_CLASS_Hour);


        /*
         * Set alerts expressed in seconds
         *
         * \note If repeat flag is set to true, alerts are triggered every x minutes
         *
         * \note If a new alert configuration is performed before the execution of the current alert,
         *       the old alert will be discarded and the new one will be taken into consideration.
         *
         * \note At any time, alerts can be disabled by calling time_alert_class_clear()
         *
         */
        time_alert_class_set(my_alert.sec, my_alert.repeat_s, SET_ALERT_CLASS_Sec);


#if (TIME_ALERT_BREATH_TIMER_EN == 1)
        /*
         * Breath timer parameters
         */
        breath_config breath_cfg = {
                .dc_min   = 0,
                .dc_max   = 255,
                .freq_div = 255,
                .dc_step  = 96,
                .polarity = HW_BREATH_PWM_POL_POS
        };
#endif

        for (;;) {

                OS_BASE_TYPE ret;
                uint32_t notif;

                /*
                 * Block forever waiting for any of the notification bits,
                 * then clear them all
                 */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS,
                                           &notif, OS_TASK_NOTIFY_FOREVER);

                /*
                 * Thus, the return value must always be OS_OK
                 */
                OS_ASSERT(ret == OS_OK);


                /*
                 * This code block is executed upon an alert event
                 * produced by Hours Alert Class.
                 */
                if (notif & TIME_ALERT_TRIG_NOTIF_H) {

#if (TIME_ALERT_BREATH_TIMER_EN == 1)
                        /* Start blinking white LED D1 on pro DevKit */
                        time_alert_led_blink_start(&breath_cfg);
#endif

                        /* Print out a debugging message */
                        if (TIME_ALERT_DBG_MESSAGES_CONSOLE) {
                                printf("The source of the alert is: 'HOURS'\n\r");
                        }
                }


                /*
                 * This code block is executed upon an alert event
                 * produced by Minutes Alert Class.
                 */
                if (notif & TIME_ALERT_TRIG_NOTIF_M) {

#if (TIME_ALERT_BREATH_TIMER_EN == 1)
                        /* Start blinking white LED D1 on pro DevKit */
                        time_alert_led_blink_start(&breath_cfg);
#endif

                        /* Print out a debugging message */
                        if (TIME_ALERT_DBG_MESSAGES_CONSOLE) {
                                printf("The source of the alert is: 'MINUTES'\n\r");
                        }
                }


                /*
                 * This code block is executed upon an alert event
                 * produced by Seconds Alert Class.
                 */
                if (notif & TIME_ALERT_TRIG_NOTIF_S) {

#if (TIME_ALERT_BREATH_TIMER_EN == 1)
                        /* Start blinking white LED D1 on pro DevKit */
                        time_alert_led_blink_start(&breath_cfg);
#endif


                        /* Print out a debugging message */
                        if (TIME_ALERT_DBG_MESSAGES_CONSOLE) {
                                printf("The source of the alert is: 'SECONDS'\n\r");
                        }

                }
        }
}



