/**
 ****************************************************************************************
 *
 * @file main.c
 *
 * @brief FreeRTOS template application with retarget
 *
 * Copyright (C) 2015 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#include <string.h>
#include <stdio.h>
#include <stdbool.h>

#include "osal.h"
#include "resmgmt.h"
#include "hw_cpm.h"
#include "hw_gpio.h"
#include "hw_watchdog.h"
#include "sys_clock_mgr.h"
#include "sys_power_mgr.h"


/*
 * Required header files for the main.c source file
 */
#include "hw_wkup.h"
#include "sw_rtc_date_time.h"
#include "alert_mechanism.h"
#include "alarm_mechanism.h"


/* Function prototypes */
extern void time_alert_led_blink_stop(void);


/* Task priorities */
#define mainRTC_TASK_PRIORITY    ( OS_TASK_PRIORITY_NORMAL )
#define mainALERT_TASK_PRIORITY  ( OS_TASK_PRIORITY_NORMAL )
#define mainALARM_TASK_PRIORITY  ( OS_TASK_PRIORITY_NORMAL )



/* The configCHECK_FOR_STACK_OVERFLOW setting in FreeRTOSConifg can be used to
check task stacks for overflows.  It does not however check the stack used by
interrupts.  This demo has a simple addition that will also check the stack used
by interrupts if mainCHECK_INTERRUPT_STACK is set to 1.  Note that this check is
only performed from the tick hook function (which runs in an interrupt context).
It is a good debugging aid - but won't catch interrupt stack problems until the
tick interrupt next executes. */
//#define mainCHECK_INTERRUPT_STACK			1
#if mainCHECK_INTERRUPT_STACK == 1
	const unsigned char ucExpectedInterruptStackValues[] = { 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC };
#endif


/*
 * Perform any application specific hardware configuration.  The clocks,
 * memory, etc. are configured before main() is called.
 */
static void prvSetupHardware( void );


static OS_TASK xHandle;


/******** Notification Bitmasks *************/
#define SW_RTC_TRIG_NOTIF   (1 << 1)



/*
 * Macro for outputting the low power (lp) clock on P3.0 pin.
 * 1: configures P3.0 pin to output the lp clock.
 * 0: nothing
 *
 * \note: This macro is intended for debugging purposes only in order to measure
 *        the accuracy of the lp clock used. The device should stay active for
 *        as long as measurements on P3.0 pin are performed.
 */
#define TEST_LP_CLOCK_OUTPUT 0


#if (SW_RTC_TASK_ENABLE == 1)
  /* Task handle */
__RETAINED_RW static OS_TASK task_rtc_h   = NULL;

  /* Function prototype */
void prvRTC_Task( void *pvParameters );
#endif


#if (ALERT_TASK_ENABLE == 1)
  /* Task handle */
__RETAINED_RW static OS_TASK task_alert_h = NULL;

   /* Function prototype */
void prvALERT_Task( void *pvParameters );
#endif


#if (ALARM_TASK_ENABLE == 1)
  /* Task handle */
__RETAINED_RW static OS_TASK task_alarm_h = NULL;

    /* Function prototype */
void prvALARM_Task( void *pvParameters );
#endif


#if (SW_RTC_TASK_ENABLE == 1)
//*************** For verification purposes - This code block will be discarded ***************
void rtc_dbg_cb(OS_TIMER xTimer)
{
     OS_TASK_NOTIFY(task_rtc_h, SW_RTC_TRIG_NOTIF, OS_NOTIFY_SET_BITS);
}
//************************************ For verification purposes *******************************
#endif


static void system_init( void *pvParameters )
{

#if defined CONFIG_RETARGET
        extern void retarget_init(void);
#endif

        /*
         * Prepare clocks. Note: cm_cpu_clk_set() and cm_sys_clk_set()
         * can be called only from a task since they will suspend the
         * task until the  XTAL16M has settled and, maybe, the PLL is
         * locked.
         */
        cm_sys_clk_init(sysclk_XTAL16M);
        cm_apb_set_clock_divider(apb_div1);
        cm_ahb_set_clock_divider(ahb_div1);
        cm_lp_clk_init();

        /* Prepare the hardware to run this demo. */
        prvSetupHardware();

        /* init resources */
        resource_init();

#if defined CONFIG_RETARGET
        retarget_init();
#endif


#if (TEST_LP_CLOCK_OUTPUT == 1)
        /* Set the device to stay in active mode. */
        pm_set_sleep_mode(pm_mode_active);

        /*
         * Select which clock to map when PID = FUNC_CLOCK.
         *
         * Valid values are:
         * [0x00]: When XTAL32K is the selected lp clock.
         * [0x02]: When RCX is the selected lp clock.
         */
        REG_SETF(GPIO, GPIO_CLK_SEL, FUNC_CLOCK_SEL, 0x00);
#else
        /* Set the desired sleep mode. */
        pm_set_sleep_mode(pm_mode_extended_sleep);
#endif


#if (SW_RTC_TASK_ENABLE == 1)
        /*
         * Task responsible for performing RTC operations
         */
        OS_TASK_CREATE( "RTC",

                        prvRTC_Task,
                        NULL,
                        300 * OS_STACK_WORD_SIZE,

                        mainRTC_TASK_PRIORITY,
                        task_rtc_h);
        OS_ASSERT(task_rtc_h);
#endif

#if (ALERT_TASK_ENABLE == 1)
        /*
         * Task responsible for performing ALERT operations
         */
        OS_TASK_CREATE( "ALERT",

                        prvALERT_Task,
                        NULL,
                        300 * OS_STACK_WORD_SIZE,

                        mainALERT_TASK_PRIORITY,
                        task_alert_h );
        OS_ASSERT(task_alert_h);
#endif


#if (ALARM_TASK_ENABLE == 1)
        /*
         *  Task responsible for performing ALARM operations
         */
        OS_TASK_CREATE( "ALARM",

                        prvALARM_Task,
                        NULL,
                        300 * OS_STACK_WORD_SIZE,

                        mainALARM_TASK_PRIORITY,
                        task_alarm_h);
        OS_ASSERT(task_alarm_h);
#endif


#if (SW_RTC_TASK_ENABLE == 1)
        //********** For verification purposes - This code block will be discarded *********
        // You can synchronize the App with your smartphone and keep track of date/time.
        OS_TIMER rtc_dbg_h;

        rtc_dbg_h = OS_TIMER_CREATE("RCX",
                                OS_MS_2_TICKS(1000), OS_TIMER_SUCCESS,
                                                   (void *) OS_GET_CURRENT_TASK(), rtc_dbg_cb);
        OS_ASSERT(rtc_dbg_h);

        /* Start the OS timer */
        OS_TIMER_START(rtc_dbg_h, OS_TIMER_FOREVER);
        //************************* For verification purposes ***************************
#endif


        /* The work of the SysInit task is done */
        OS_TASK_DELETE( xHandle );

}

/*
 * Callback function that is called upon a WKUP event. In our case
 * after K1 button on the Pro DevKit is pressed.
 */
void wkup_cb(void)
{
        /*
         * This function must be called by any user-defined
         * interrupt handler, to clear the interrupt flag.
         */
        hw_wkup_reset_interrupt();

#if (SW_RTC_TASK_ENABLE == 1)
       /*
        * Notify the main task that time for performing RTC operations has elapsed
        */
        OS_TASK_NOTIFY_FROM_ISR(task_rtc_h, SW_RTC_TRIG_NOTIF,
                                              OS_NOTIFY_SET_BITS);
#endif

        /*
         * Disable Breath timer and resume the device to its
         * initial state, that is extended sleep mode!
         */
        time_alert_led_blink_stop();

}


/**
 * @brief Template main creates a SysInit task, which creates a Template task
 */
int main( void )
{
        OS_BASE_TYPE status;

        cm_clk_init_low_level();                            /* Basic clock initializations. */

        /* Start the two tasks as described in the comments at the top of this
        file. */
        status = OS_TASK_CREATE("SysInit",              /* The text name assigned to the task, for
                                                           debug only; not used by the kernel. */
                        system_init,                    /* The System Initialization task. */
                        ( void * ) 0,                   /* The parameter passed to the task. */
                        configMINIMAL_STACK_SIZE * OS_STACK_WORD_SIZE,
                                                        /* The number of bytes to allocate to the
                                                           stack of the task. */
                        OS_TASK_PRIORITY_HIGHEST,       /* The priority assigned to the task. */
                        xHandle );                      /* The task handle */
        OS_ASSERT(status == OS_TASK_CREATE_SUCCESS);


	/* Start the tasks and timer running. */
	vTaskStartScheduler();

	/* If all is well, the scheduler will now be running, and the following
	line will never be reached.  If the following line does execute, then
	there was insufficient FreeRTOS heap memory available for the idle and/or
	timer tasks	to be created.  See the memory management section on the
	FreeRTOS web site for more details. */
	for( ;; );

}


/*
 * Function which makes all the necessary initializations for the
 * wake-up controller
 */
static void init_wkup(void)
{
        /*
         * This function must be called first and is responsible
         * for the initialization of the WKUP hardware block.
         */
        hw_wkup_init(NULL);

        /*
         * Configure the pin(s) that can trigger the device to
         * wake up while in sleep mode. The last input parameter
         * determines the triggering edge of the pulse (event)
         */
        hw_wkup_configure_pin(HW_GPIO_PORT_1, HW_GPIO_PIN_6,
                                           true, HW_WKUP_PIN_STATE_LOW);

        /*
         * This function defines a delay between the moment at which a trigger event
         * is present on a pin and the moment at which the controller takes this
         * event into consideration. Setting debounce time to 0, hardware debouncing
         * mechanism is disabled. Maximum debounce time is 63 ms.
         */
        hw_wkup_set_debounce_time(10);

// Check if the chip is either DA14680/81 or not
#if dg_configBLACK_ORCA_IC_REV == BLACK_ORCA_IC_REV_A

        /*
         * Set threshold for event counter. Interrupt is generated after
         * the event counter reaches the configured value. This function
         * is only supported in DA14680/1 chips.
         */
        hw_wkup_set_counter_threshold(1);
#endif

        /* Register interrupt handler */
        hw_wkup_register_interrupt(wkup_cb, 1);
}



/**
 * @brief Initialize the peripherals domain after power-up.
 *
 */
static void periph_init(void)
{
#       if dg_configBLACK_ORCA_MB_REV == BLACK_ORCA_MB_REV_D
#               define UART_TX_PORT    HW_GPIO_PORT_1
#               define UART_TX_PIN     HW_GPIO_PIN_3
#               define UART_RX_PORT    HW_GPIO_PORT_2
#               define UART_RX_PIN     HW_GPIO_PIN_3
#       else
#               error "Unknown value for dg_configBLACK_ORCA_MB_REV!"
#       endif


        hw_gpio_set_pin_function(UART_TX_PORT, UART_TX_PIN,
                       HW_GPIO_MODE_OUTPUT, HW_GPIO_FUNC_UART_TX);
        hw_gpio_set_pin_function(UART_RX_PORT, UART_RX_PIN,
                       HW_GPIO_MODE_INPUT, HW_GPIO_FUNC_UART_RX);

#if (TEST_LP_CLOCK_OUTPUT == 1)
        // Output the selected lp clock on P3.0 pin - Intended for debugging purposes only.
        hw_gpio_set_pin_function(HW_GPIO_PORT_3, HW_GPIO_PIN_0, HW_GPIO_MODE_OUTPUT,
                        HW_GPIO_FUNC_CLOCK);
#endif

}

/**
 * @brief Hardware Initialization
 */
static void prvSetupHardware( void )
{
        /* Init hardware */
        pm_system_init(periph_init);
        init_wkup();
}


/**
 * @brief Malloc fail hook
 */
void vApplicationMallocFailedHook( void )
{
	/* vApplicationMallocFailedHook() will only be called if
	configUSE_MALLOC_FAILED_HOOK is set to 1 in FreeRTOSConfig.h.  It is a hook
	function that will get called if a call to OS_MALLOC() fails.
	OS_MALLOC() is called internally by the kernel whenever a task, queue,
	timer or semaphore is created.  It is also called by various parts of the
	demo application.  If heap_1.c or heap_2.c are used, then the size of the
	heap available to OS_MALLOC() is defined by configTOTAL_HEAP_SIZE in
	FreeRTOSConfig.h, and the xPortGetFreeHeapSize() API function can be used
	to query the size of free heap space that remains (although it does not
	provide information on how the remaining heap might be fragmented). */
	taskDISABLE_INTERRUPTS();
	for( ;; );
}

/**
 * @brief Application idle task hook
 */
void vApplicationIdleHook( void )
{
	/* vApplicationIdleHook() will only be called if configUSE_IDLE_HOOK is set
	to 1 in FreeRTOSConfig.h.  It will be called on each iteration of the idle
	task.  It is essential that code added to this hook function never attempts
	to block in any way (for example, call OS_QUEUE_GET() with a block time
	specified, or call OS_DELAY()).  If the application makes use of the
	OS_TASK_DELETE() API function (as this demo application does) then it is also
	important that vApplicationIdleHook() is permitted to return to its calling
	function, because it is the responsibility of the idle task to clean up
	memory allocated by the kernel to any task that has since been deleted. */
}

/**
 * @brief Application stack overflow hook
 */
void vApplicationStackOverflowHook( OS_TASK pxTask, char *pcTaskName )
{
	( void ) pcTaskName;
	( void ) pxTask;

	/* Run time stack overflow checking is performed if
	configCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook
	function is called if a stack overflow is detected. */
	taskDISABLE_INTERRUPTS();
	for( ;; );
}

/**
 * @brief Application tick hook
 */
void vApplicationTickHook( void )
{
#if mainCHECK_INTERRUPT_STACK == 1
extern unsigned long _pvHeapStart[];

	/* This function will be called by each tick interrupt if
	configUSE_TICK_HOOK is set to 1 in FreeRTOSConfig.h.  User code can be
	added here, but the tick hook is called from an interrupt context, so
	code must not attempt to block, and only the interrupt safe FreeRTOS API
	functions can be used (those that end in FromISR()). */

	/* Manually check the last few bytes of the interrupt stack to check they
	have not been overwritten.  Note - the task stacks are automatically
	checked for overflow if configCHECK_FOR_STACK_OVERFLOW is set to 1 or 2
	in FreeRTOSConifg.h, but the interrupt stack is not. */
	OS_ASSERT( memcmp( ( void * ) _pvHeapStart, ucExpectedInterruptStackValues, sizeof( ucExpectedInterruptStackValues ) ) == 0U );
#endif /* mainCHECK_INTERRUPT_STACK */
}





