/**
 ****************************************************************************************
 *
 * @file sw_rtc_date_time.c
 *
 * @brief SW RTC mechanism implementation
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#include "sw_rtc_date_time.h"

#include <stdbool.h>
#include <stdlib.h>

#include <math.h>
#include "osal.h"

#include "sys_rtc.h"
#include "sys_power_mgr.h"

#include <string.h>
#include <stdio.h>



/*
 * Set this macro to perform a forced RTC measurement via an OS Timer.
 */
#define SW_RTC_FORCED_DATE_TIME_EN    1


/* Mutex handle */
__RETAINED_RW static OS_MUTEX sw_rtc_mutex = NULL;


/*
 * Time interval, expressed in milliseconds, for triggering a forced RTC operation.
 * This macro is used only when SW_RTC_FORCED_DATE_TIME_EN is set. Default value is
 * 1 day, that is:
 *
 * (24 hrs * 60 minutes/hour * 60 seconds/minute * 1000 milliseconds/second) = 86,400,000
 *
 */
#define SW_RTC_FORCED_DATE_TIME_IN_MS   86400000


/*
 * Variables used to store the lp clocks, using the old/new scheme
 */
__RETAINED_RW static uint64_t sw_rtc_new = 0;
__RETAINED_RW static uint64_t sw_rtc_old = 0;


/*
 * SW RTC parameters. This structure holds the current date/time.
 */
__RETAINED_RW static sw_rtc_date_time_t *calendar;


/* RTC mechanism status flag */
__RETAINED_RW static bool sw_rtc_status_flag = false;


#if (SW_RTC_FORCED_DATE_TIME_EN == 1)

/* OS timer handle */
static OS_TIMER sw_rtc_timer_h;


/* Callback function called after OS timer's expiration */
void sw_rtc_trg_cb(OS_TIMER timer)
{
        OS_BASE_TYPE ret;

        /*
         * Try to get the mutex without blocking. If the mutex was successfully acquired then
         * no other tasks perform RTC measurements. If not, another task is already performing
         * RTC operations and it's OK to abort the operation.
         *
         * \warning Blocking within timer's callback is not permitted!!!
         *
         */
        ret = OS_MUTEX_GET(sw_rtc_mutex, OS_MUTEX_NO_WAIT);

        if (ret == OS_MUTEX_TAKEN) {
                /* Allocate memory both date/time */
                sw_rtc_date_time_t *date_time = (sw_rtc_date_time_t *) OS_MALLOC(sizeof(sw_rtc_date_time_t));

                /* Just perform a read date/time operation */
                sw_rtc_date_time_read(date_time);

                /* Release the previously allocated memory */
                OS_FREE(date_time);

                /* Release the already acquired mutex */
                OS_MUTEX_PUT(sw_rtc_mutex);

        }
}

#endif


/*
 * Initialize the SW RTC mechanism. First, memory for the SW RTC parameters
 * is allocated and then an OS timer responsible for performing forced RTC
 * operations is created.
 *
 */
void sw_rtc_init(void)
{
        /* Allocate memory for the calendar (date/time) structure */
        calendar = (sw_rtc_date_time_t *) OS_MALLOC(sizeof(sw_rtc_date_time_t));
        OS_ASSERT(calendar);

        /* Clear the calendar structure */
        memset((uint8_t *)calendar, 0x00, sizeof(sw_rtc_date_time_t));


#if (SW_RTC_FORCED_DATE_TIME_EN == 1)

        uint64_t timer_ticks = (((uint64_t)SW_RTC_FORCED_DATE_TIME_IN_MS * configTICK_RATE_HZ) / 1000);

        /*
         * Create an OS timer to trigger a forced RTC operation.
         */
        sw_rtc_timer_h = OS_TIMER_CREATE("RTC_Trigger",
                                (uint32_t)timer_ticks, OS_TIMER_SUCCESS,
                                                   (void *) OS_GET_CURRENT_TASK(), sw_rtc_trg_cb);
        OS_ASSERT(sw_rtc_timer_h);

        /* Start the OS timer */
        OS_TIMER_START(sw_rtc_timer_h, OS_TIMER_FOREVER);
#endif

        /* Create mutex for RTC operations */
        OS_MUTEX_CREATE(sw_rtc_mutex);
        OS_ASSERT(sw_rtc_mutex != NULL);

}


/* Reset calendar values to zero */
void sw_rtc_date_time_reset(void)
{
        /* Reset all the calendar related parameters */
        sw_rtc_new = 0;
        sw_rtc_old = 0;

        /* Reset status flag */
        sw_rtc_status_flag = false;

        /* Reset date parameters */
        calendar->date.year       = 0;
        calendar->date.month      = 1;
        calendar->date.month_date = 1;

        /* Reset time parameters */
        calendar->time.hour = 0;
        calendar->time.min  = 0;
        calendar->time.sec  = 0;
        calendar->time.msec = 0;
        calendar->time.usec = 0;
}


/*
 * Get the number of days of a month.
 */
uint8_t sw_rtc_get_days_of_month(void)
{
        uint8_t days = 0;

        /* Given the current month, calculate its maximum days */
        switch(calendar->date.month) {

        case JANUARY:
        case MARCH:
        case MAY:
        case JULY:
        case AUGUST:
        case OCTOBER:
        case DECEMBER:
                days = 31;
                break;
        case FEBRUARY:
                /*
                 * Check if the year is a leap year, taking into consideration
                 * the following three criteria:
                 * 1. the year can be evenly divided by 4
                 * 2. the year is divisible by 400
                 * 3. the year is not divisible by 100
                 */
                if ((!(calendar->date.year % 4)) ||
                        ((!(calendar->date.year % 400)) && (calendar->date.year % 100))) {
                        days = 29;
                } else {
                        days = 28;
                }
                break;
        case APRIL:
        case JUNE:
        case SEPTEMBER:
        case NOVEMBER:
                days = 30;
                break;
        default:
                OS_ASSERT(0); // Invalid value!
        }

        return days;
}




/* Check whether the given year is a leap year */
bool sw_rtc_is_leap_year(void)
{
        /*
         * Check if the year is a leap year, taking into consideration
         * the following three criteria:
         * 1. the year can be evenly divided by 4
         * 2. the year is divisible by 400
         * 3. the year is not divisible by 100
         */
        if ((!(calendar->date.year % 4)) ||
                ((!(calendar->date.year % 400)) && (calendar->date.year % 100))) {
                return true;
        } else {
                return false;
        }
}


/* Calculate the month code used in weekday computations */
uint8_t sw_rtc_compute_month_code(void)
{
        uint8_t month_code;

        switch(calendar->date.month) {

        case JANUARY:
                month_code = 0;
                break;
        case FEBRUARY:
                month_code = 3;
                break;
        case MARCH:
                month_code = 3;
                break;
        case APRIL:
                month_code = 6;
                break;
        case MAY:
                month_code = 1;
                break;
        case JUNE:
                month_code = 4;
                break;
        case JULY:
                month_code = 6;
                break;
        case AUGUST:
                month_code = 2;
                break;
        case SEPTEMBER:
                month_code = 5;
                break;
        case OCTOBER:
                month_code = 0;
                break;
        case NOVEMBER:
                month_code = 3;
                break;
        case DECEMBER:
                month_code = 5;
                break;
        default:
                OS_ASSERT(0); // Invalid value!
        }

        return month_code;
}

/* Calculate the century code used in weekday computations */
uint8_t sw_rtc_compute_century_code(void)
{
        uint8_t century_code = 0;

        if (calendar->date.year >= 2000 && calendar->date.year < 2100) {
                century_code = 6;
        } else if (calendar->date.year >= 2100 && calendar->date.year < 2200) {
                century_code = 4;
        } else if (calendar->date.year >= 2200 && calendar->date.year < 2300) {
                century_code = 2;
        } else if (calendar->date.year >= 2300 && calendar->date.year < 2400) {
                century_code = 0;
        }

        return century_code;
}

/* Function responsible for computing the current day of the week */
uint32_t sw_rtc_compute_date_of_week(void)
{
      /* Isolate the last two digits from the year e.g. 2018 -> 18 */
      uint8_t year_code = (uint8_t)(calendar->date.year % 100);

      /* Compute the year code */
      year_code = ((year_code + (year_code / 4)) % 7);

      /* Compute the month code */
      uint8_t month_code = sw_rtc_compute_month_code();

      /* Compute the century code */
      uint8_t centure_code = sw_rtc_compute_century_code();

      /* Check whether the year is a leap year */
      uint8_t leap_year = (uint8_t)sw_rtc_is_leap_year();

      /* Add all the codes */
      uint8_t day_week = (year_code + month_code + centure_code + (uint8_t)calendar->date.month_date);


      if (leap_year && ((calendar->date.month == JANUARY) || (calendar->date.month == FEBRUARY)) ) {
              return (uint32_t)((day_week - 1) % 7);
      } else {
              return (uint32_t)(day_week % 7);
      }

}



/*
 * Function responsible for converting time, expressed in microseconds (us), into full calendar (date/time).
 */
void sw_rtc_convert_time_to_calendar(uint64_t usec_time)
{
        calendar->time.usec += usec_time;

        /*
         * Check if there is an overflow in [microseconds],
         * that is greater than 999.
         */
        if (calendar->time.usec > 999) {
                /* Compute the right next greater unit, that is milliseconds */
                calendar->time.msec += (uint32_t)(calendar->time.usec / 1000);

                /* Then compute the new value for [microseconds] */
                calendar->time.usec %= 1000;

                /*
                 * Check if there is an overflow in [milliseconds],
                 * that is greater than 999.
                 */
                if (calendar->time.msec > 999) {
                        /* Compute the right next greater unit, that is [seconds] */
                        calendar->time.sec += (calendar->time.msec / 1000);

                        /* Then compute the new value for [milliseconds] */
                        calendar->time.msec %= 1000;

                        /*
                         * Check if there is an overflow in [seconds],
                         * that is greater than 59
                         */
                        if (calendar->time.sec > 59) {
                                /* Compute the right next greater unit, that is [minutes] */
                                calendar->time.min += (calendar->time.sec / 60);

                                /* Then compute the new value for [seconds] */
                                calendar->time.sec %= 60;

                                /*
                                 * Check is there is an overflow in [minutes],
                                 * that is greater than 59
                                 */
                                if (calendar->time.min > 59) {
                                        /* Compute the right next greater unit, that is [hours] */
                                        calendar->time.hour += (calendar->time.min / 60);

                                        /* Then compute the new value for [minutes] */
                                        calendar->time.min %= 60;

                                        /*
                                         * Check if there is an overflow in hours,
                                         * that is greater than 23
                                         */
                                        if (calendar->time.hour > 23) {
                                                /* Compute the right next greater unit, that is [days] */
                                                calendar->date.month_date += (calendar->time.hour / 24);

                                                /* Then compute the new value for [hours] */
                                                calendar->time.hour %= 24;

                                                /*
                                                 * Check if there is an overflow in days,
                                                 * that is greater than 28/29/30 or 31
                                                 * depending on the current month.
                                                 */
                                                if (calendar->date.month_date >
                                                                sw_rtc_get_days_of_month()) {

                                                        /* Calculate months and days */
                                                        do {
                                                                calendar->date.month_date -= sw_rtc_get_days_of_month();
                                                                calendar->date.month++;

                                                                /*
                                                                 * Check if there is an overflow in [months],
                                                                 * that is greater than 12
                                                                 */
                                                                if (calendar->date.month > 12) {
                                                                        calendar->date.year += (calendar->date.month / 12);
                                                                        calendar->date.month %= 12;
                                                                }

                                                        } while (calendar->date.month_date > sw_rtc_get_days_of_month());


                                                } // end of days monitoring
                                        } // end of hours monitoring
                                } // end of minutes monitoring
                        } // end of seconds monitoring
                } // end of milliseconds monitoring
        } // end of microseconds monitoring

        /* Finally compute the day of the week */
        calendar->date.week_date = sw_rtc_compute_date_of_week();

} // end of function




/*
 * Convert low power (lp) clock cycles into time expressed in microseconds.
 */
uint64_t sw_rtc_convert_lp_to_time(uint64_t lp_clocks)
{
        uint64_t time = 0;

        /*
         * Check whether the lp clock source is either the external crystal XTAL32K
         * or an external clock source.
         */
        if ((dg_configUSE_LP_CLK == LP_CLK_32768) ||
                             (dg_configUSE_LP_CLK == LP_CLK_32000)) {


                /*
                 * Compute the period of the lp clock used
                 * and then multiply it with 1024 to increase
                 * accuracy!
                 */
                const uint32_t lp_clk_period = ((1000000UL * 1024)
                                                /configSYSTICK_CLOCK_HZ);

                /*
                 * Convert lp clocks into time
                 */
                time = (lp_clocks * (uint64_t)lp_clk_period);
                time = (time >> 10); // divide with 1024

       /* Check whether the lp clock source is the internal RCX */
        } else if (dg_configUSE_LP_CLK == LP_CLK_RCX) {

                /*
                 * Use [rcx_clock_period] to get the current RCX period in usec.
                 * Please note that this value is multiplied by [1024 * 1024]
                 */
                time = (lp_clocks * (uint64_t)rcx_clock_period);
                time = (time >> 20); // divide with (1024 * 1024)

        } else {

                OS_ASSERT(0); // Invalid lp clock source
        }

        return time;
}



/*
 * \callgraph
 *
 * \brief This function performs a SW RTC measurement using APIs from the RTC mechanism
 *        (provided by the SDK), to calculate the number of low power clock cycles that
 *        fit within a given time interval.
 */
void sw_rtc_date_time_read(sw_rtc_date_time_t *val)
{

        uint64_t sw_rtc_value_us = 0;
        uint64_t sw_rtc_dif = 0;


        /* Get the current RTC value expressed in lp clock ticks. */
        sw_rtc_new = rtc_get();


        /*
         * To reduce values used in various calculations, calculate
         * only the lp clocks since the last RTC measurement.
         * (instead of invoking all the lp clock cycles measured
         * from the very beginning of chip's operation)
         */
        sw_rtc_dif = sw_rtc_new - sw_rtc_old;


        /*
         * The current RTC value becomes the old  for the next RTC measurement.
         */
        sw_rtc_old = sw_rtc_new;


        /* Convert the lp clock cycles into [microseconds]  */
        sw_rtc_value_us = sw_rtc_convert_lp_to_time(sw_rtc_dif);


        /* Update the calendar using the current time in [microseconds]  */
        sw_rtc_convert_time_to_calendar(sw_rtc_value_us);


        /* Safeguard the structure that holds the current date/time */
        OS_MUTEX_GET(sw_rtc_mutex, OS_MUTEX_FOREVER);

        /* Copy the current date/time */
        memcpy((uint8_t *)val, (uint8_t *)calendar, sizeof(sw_rtc_date_time_t));

        /* Release the previously acquired mutex so that other tasks can use it e.g the daemon task */
        OS_MUTEX_PUT(sw_rtc_mutex);
}


/*
 * Function that performs an SW RTC measurement and returns the current date.
 */
void sw_rtc_date_read(sw_rtc_date_t *val)
{

        /* Allocate memory to hold both date and time */
        sw_rtc_date_time_t *date_time = (sw_rtc_date_time_t *) OS_MALLOC(sizeof(sw_rtc_date_time_t));
        OS_ASSERT(date_time);

        /* Read the current date/time */
        sw_rtc_date_time_read(date_time);

        /* Safeguard the structure that holds the current date/time */
        OS_MUTEX_GET(sw_rtc_mutex, OS_MUTEX_FOREVER);

        /* Copy the current date */
        memcpy((uint8_t *)val, (uint8_t *)&date_time->date, sizeof(sw_rtc_date_t));

        /* Release the previously acquired mutex so that other tasks can use it e.g the daemon task */
        OS_MUTEX_PUT(sw_rtc_mutex);


        OS_FREE(date_time);
}



/*
 * Function that performs an SW RTC measurement and returns the current time.
 */
void sw_rtc_time_read(sw_rtc_time_t *val)
{

        /* Allocate memory to hold both date and time */
        sw_rtc_date_time_t *date_time = (sw_rtc_date_time_t *) OS_MALLOC(sizeof(sw_rtc_date_time_t));
        OS_ASSERT(date_time);

        /* Read the current date/time */
        sw_rtc_date_time_read(date_time);

        /* Safeguard the structure that holds the current date/time */
        OS_MUTEX_GET(sw_rtc_mutex, OS_MUTEX_FOREVER);

        /* Copy the current time  */
        memcpy((uint8_t *)val, (uint8_t *)&date_time->time, sizeof(sw_rtc_time_t));

        /* Release the previously acquired mutex so that other tasks can use it e.g the daemon task */
        OS_MUTEX_PUT(sw_rtc_mutex);


        OS_FREE(date_time);
}


/* Sanity checks for the date structure */
int sw_rtc_date_write_sanity_check(sw_rtc_date_t *val)
{
        if ((val->month_date > 31) || (val->month_date == 0) || (val->month > 12) || (val->month == 0) || (val->week_date > 6)) {
                return -1;
        } else {
                return 0;
        }
}


/* Sanity checks for the time structure */
int sw_rtc_time_write_sanity_check(sw_rtc_time_t *val)
{
        if ((val->hour > 23) || (val->min > 59) || (val->sec > 59) || (val->msec > 999) || (val->usec > 999))  {
                return -1;
        } else {
                return 0;
        }
}


/*
 * Function responsible for setting the current date and time
 */
void sw_rtc_date_time_write(sw_rtc_date_time_t *val)
{
        int error_date;
        int error_time;


        /**
         * Perform sanity checks on the provided values.
         *
         * \note If the returned value is -1 then the provided values are out of the boundaries
         *
         */
        error_date = sw_rtc_date_write_sanity_check(&val->date);
        error_time = sw_rtc_time_write_sanity_check(&val->time);

        OS_ASSERT((error_date == 0) && (error_time == 0));


        /* Safeguard the structure that holds the current date/time */
        OS_MUTEX_GET(sw_rtc_mutex, OS_MUTEX_FOREVER);

        /* Update date/time parameters */
        memcpy((uint8_t *)calendar, (uint8_t *)val, sizeof(sw_rtc_date_time_t));

        /* Release the previously acquired mutex so that other tasks can use it e.g the daemon task */
        OS_MUTEX_PUT(sw_rtc_mutex);

        /* SW RTC mechanism is up and running */
        sw_rtc_status_flag = true;
}




/*
 * Function responsible for setting the current date
 */
void sw_rtc_date_write(sw_rtc_date_t *val)
{
        int error;


        /**
         * Perform sanity checks on the provided values.
         *
         * \note If the returned value is -1 then the provided values are out of the boundaries
         *
         */
        error = sw_rtc_date_write_sanity_check(val);
        OS_ASSERT(error == 0);

        /* Safeguard the structure that holds the current date */
        OS_MUTEX_GET(sw_rtc_mutex, OS_MUTEX_FOREVER);

        /* Update date parameters */
        memcpy((uint8_t *)&calendar->date, (uint8_t *)val, sizeof(sw_rtc_date_t));

        /* Release the previously acquired mutex so that other tasks can use it e.g the daemon task */
        OS_MUTEX_PUT(sw_rtc_mutex);

        /* SW RTC mechanism is up and running */
        sw_rtc_status_flag = true;
}


/*
 * Function responsible for setting the current time
 */
void sw_rtc_time_write(sw_rtc_time_t *val)
{
        int error;


        /**
         * Perform sanity checks on the provided values.
         *
         * \note If the returned value is -1 then the provided values are out of  boundaries
         *
         */
        error = sw_rtc_time_write_sanity_check(val);
        OS_ASSERT(error == 0);


        /* Safeguard the structure that holds the current time */
        OS_MUTEX_GET(sw_rtc_mutex, OS_MUTEX_FOREVER);

        /* Update time parameters */
        memcpy((uint8_t *)&calendar->time, (uint8_t *)val, sizeof(sw_rtc_time_t));

        /* Release the previously acquired mutex so that other tasks can use it e.g the daemon task */
        OS_MUTEX_PUT(sw_rtc_mutex);

        /* SW RTC mechanism is up and running */
        sw_rtc_status_flag = true;
}


/* Get the RTC mechanism status */
bool sw_rtc_check_status(void)
{
        return sw_rtc_status_flag;
}
