/**
 ****************************************************************************************
 *
 * @file sw_rtc_date_time.h
 *
 * @brief SW RTC mechanism APIs
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#ifndef sw_rtc_DATE_TIME_H_
#define sw_rtc_DATE_TIME_H_

#include <stdint.h>
#include <stdbool.h>


/* Days of a week */
typedef enum {
        WEEK_DATE_SUNDAY     = 0,
        WEEK_DATE_MONDAY     = 1,
        WEEK_DATE_TUESDAY    = 2,
        WEEK_DATE_WEDNESDAY  = 3,
        WEEK_DATE_THURSDAY   = 4,
        WEEK_DATE_FRIDAY     = 5,
        WEEK_DATE_SATURDAY   = 6,
} WEEK_DATE;



/* Months of a year */
typedef enum  {
        JANUARY   = 1,
        FEBRUARY  = 2,
        MARCH     = 3,
        APRIL     = 4,
        MAY       = 5,
        JUNE      = 6,
        JULY      = 7,
        AUGUST    = 8,
        SEPTEMBER = 9,
        OCTOBER   = 10,
        NOVEMBER  = 11,
        DECEMBER  = 12
} SW_RTC_MONTH;


/*
 * Date related parameters
 */
typedef struct sw_rtc_date {
        uint32_t year;       // Current year: there is no limitation
        uint32_t month;      // Range: (1..12)
        uint32_t month_date; // Range: (1..31) or (1..30) or (1..29) or (1..28)
        uint32_t week_date;  // Range: (0:6) where 0 = Sunday ... 6 = Saturday
} sw_rtc_date_t;


/*
 * Time related parameters
 */
typedef struct sw_rtc_time {
        uint32_t hour;  // Range: (0..23)
        uint32_t min;   // Range: (0..59)
        uint32_t sec;   // Range: (0..59)
        uint32_t msec;  // Range: (0..999)
        uint64_t usec;  // Range: (0..999)
} sw_rtc_time_t;


/*
 * Date/time (calendar) related parameters
 */
typedef struct sw_rtc_calendar {
        sw_rtc_date_t date;
        sw_rtc_time_t time;
} sw_rtc_date_time_t;



/**
 *
 * \brief SW RTC initialization
 *
 * This function initializes all the resources required for the SW RTC mechanism.
 *
 * \note It should be the first invoked API before any other SW RTC related operations.
 *
 */
void sw_rtc_init();


/**
 *
 * \brief SW RTC date/time read operation
 *
 * This function is responsible for reading the current date and time. It should be
 * considered as a shortcut function since it combines both sw_rtc_date_read()
 * and sw_rtc_time_read().
 *
 *
 * \param [out] val  A structure in which the current date/time will be stored
 *
 * \note This function can be called by any task at any time, to read the current date.
 *       A protection mechanism against multiple accesses has been implemented.
 *
 */
void sw_rtc_date_time_read(sw_rtc_date_time_t *val);




/*
 * \brief SW RTC date read operation
 *
 * This function is responsible for reading the current date.
 *
 *
 * \param [out] val   A structure in which the current date will be stored
 *
 * \note This function can be called by any task at any time, to read the current date.
 *       A protection mechanism against multiple accesses has been implemented.
 *
 */
void sw_rtc_date_read(sw_rtc_date_t *val);


/*
 * \brief SW RTC time read operation
 *
 * This function is responsible for reading the current time.
 *
 *
 * \param [out] val  A structure in which the current time will be stored
 *
 * \note This function can be called by any task at any time, to read the current date.
 *       A protection mechanism against multiple accesses has been implemented.
 *
 */
void sw_rtc_time_read(sw_rtc_time_t *val);



/*
 * \brief SW RTC date/time write operation
 *
 * This function is responsible for setting the current date and time. It should be
 * considered as a shortcut function since it combines both sw_rtc_date_write()
 * and sw_rtc_time_write().
 *
 *
 * \param[in] val The preferred values for date/time.
 *
 * \note The weekday is calculated automatically by the app.
 *
 * \note The code performs sanity checks to identify values out of boundaries.
 *       For any invalid value, the code gets stuck in assertions!
 *
 * \note This function can be called by any task at any time, to set date/time.
 *       A protection mechanism against multiple accesses has been implemented.
 *
 */
void sw_rtc_date_time_write(sw_rtc_date_time_t *val);



/*
 * \brief SW RTC date write operation
 *
 * This function is responsible for setting the current time.
 *
 *
 * \param[in] val The preferred values for time
 *
 * \note The code performs sanity checks to identify values out of boundaries.
 *       For any invalid value, the code gets stuck in assertions!
 *
 * \note This function can be called by any task at any time, to set date/time.
 *       A protection mechanism against multiple accesses has been implemented.
 *
 */
void sw_rtc_time_write(sw_rtc_time_t *val);



/*
 * \brief SW RTC time write operation
 *
 * This function responsible for setting the current time.
 *
 *
 * \param[in] val The preferred values for date.
 *
 * \note The weekday is calculated automatically by the code.
 *
 * \note The code performs sanity checks to identify values out of boundaries.
 *       For any invalid value, the code gets stuck in assertions!
 *
 * \note This function can be called by any task at any time, to set date/time.
 *       A protection mechanism against multiple accesses has been implemented.
 *
 */
void sw_rtc_date_write(sw_rtc_date_t *val);



/*
 * \brief Get the SW RTC mechanism status
 *
 * This function should be used to check whether the SW RTC mechanism is
 * up and running. This implies that sw_rtc_init() and one of the APIs
 * responsible for setting the current date/time have been called.
 *
 * \return '1' if the RTC mechanism is up and running, '0' otherwise.
 *
 *
 * \note This function should be invoked before setting alarm events to check
 *       whether the readiness of the SW RTC mechanism.
 *
 *
 * \code{.c}
 *
 * ...
 *
 * while (!sw_rtc_check_status()) ;
 *
 * ...
 *
 * \endcode
 *
 */
bool sw_rtc_check_status(void);

#endif /* sw_rtc_DATE_TIME_H_ */
