/**
 ****************************************************************************************
 *
 * @file sw_rtc_task.c
 *
 * @brief SW RTC application task
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#include <string.h>
#include <stdio.h>

#include "sw_rtc_date_time.h"
#include "osal.h"


/******** Notification Bitmasks *************/
#define SW_RTC_TRIG_NOTIF        (1 << 1)


/*
 * Macro for enabling/disabling debugging messages on the serial console:
 *
 * 1: enables  debugging messages
 * 0: disables debugging messages
 */
#define SW_RTC_DBG_MESSAGES_CONSOLE   1


 /*
  * Structure that holds date/time related info
  */
__RETAINED_RW sw_rtc_date_time_t date_time = {
        /*
         * Declare all the info related to date.
         *
         * \note: The weekday is calculated automatically by the app.
         *
         * \note: The app performs sanity checks to identify values out of boundaries.
         *        For any invalid value, the code gets stuck in assertions!
         *
         */
        .date = {
                .year       = 2018,
                .month      = 10,
                .month_date = 1,
        },
        /*
         * Declare all the info related to time.
         *
         * \note: The max time resolution is microseconds (us) */
        .time = {
                .hour = 0,
                .min  = 0,
                .sec  = 0,
        }
};



/*
 * Task responsible for performing RTC operations
 */
void prvRTC_Task( void *pvParameters )
{
        char calendar_str[60];

        /*
         * Initialize the RTC mechanism. This should be the first invoked
         * function before any other related SW RTC API.
         */
        sw_rtc_init();


        /* Set the current date */
        sw_rtc_date_write(&date_time.date);

        /* Set the current time */
        sw_rtc_time_write(&date_time.time);

        /*
         * Instead of declaring date/time individually, sw_rtc_date_time_write()
         * can be used to set both of them at once:
         */


        memset((uint8_t *)&date_time, 0x00, sizeof(sw_rtc_date_time_t));

        for(;;) {

                OS_BASE_TYPE ret;
                uint32_t notif;

                /*
                 * Block forever waiting for any of the notification bits,
                 * then clear them all. In this example at every K1 button
                 * press (on pro DevKit)
                 */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS,
                                                  &notif, OS_TASK_NOTIFY_FOREVER);

                /*
                 * Thus, the return value must always be OS_OK
                 */
                OS_ASSERT(ret == OS_OK);

                /*
                 * Has time for updating the calendar elapsed?
                 */
                if (notif & SW_RTC_TRIG_NOTIF) {

                      /* Read the current date */
                      sw_rtc_date_read(&date_time.date);

                      /* Read the current time */
                      sw_rtc_time_read(&date_time.time);


                      /*
                       * Instead of reading date/time individually, sw_rtc_date_time_read()
                       * can be used to read both of them at once:
                       */


                      /* If enabled, print the current date/time on the serial console */
                      if (SW_RTC_DBG_MESSAGES_CONSOLE) {

                              /* Clear the buffer */
                              memset(calendar_str, '\0', sizeof(calendar_str));

                              /*
                               * Prepare the debugging message (current date/time)
                               */
                              sprintf(calendar_str, "DATE: %lu-%lu-%lu (week day: %lu) TIME: %lu-%lu-%lu-%lu\n\r",
                                      date_time.date.month_date, date_time.date.month, date_time.date.year, date_time.date.week_date,
                                      date_time.time.hour,       date_time.time.min,   date_time.time.sec,  date_time.time.msec );


                              /* Send the message on the serial console */
                              printf("%s", calendar_str);
                              fflush(stdout);
                      }

                }

      } // end of main for(;;) loop
} // end of task

